# Break It! example game for Jamming with Python
# programming and computer science history course content.
# (C) Lotus Fox LLC, Aubrey Smith, 2017

import pygame
from pygame.locals import *
import pygame.freetype

pygame.init()

from engine import *

# displays the main menu and tests for clicks on its buttons.
# edit this to change the contents of the main menu
class MainMenu:
  # set empty attributes, so we can keep track of what to fill
  # we'll fill these with UIElements on init
  title = None
  play_button = None
  quit_button = None
  bg = (0,0,0) # background color of entire menu
  clicking = None
  # initialize fonts; if None is given instead of a filename, uses default font.
  # second argument is size of font to use.
  # Here we're using the Roboto font by Christian Robertson et al,
  #   provided through Google Fonts.
  title_font = pygame.freetype.Font("font/Roboto/Roboto-black.ttf", 64)
  button_font = pygame.freetype.Font("font/Roboto/Roboto-regular.ttf", 32)
  credit_font = button_font
  
  def __init__(self, state):
    # "state" attribute should be the engine state.
    
    # Color format is (r, g, b) with values from 0 to 255.
    # You can also include a fourth element to the tuple for "alpha";
    #   255 is opaque and 0 is transparent. Without it, always opaque.
    fg = self.fg = (255,255,255) # white for foreground color
    bg = self.bg = (0,0,0)       # black for background color
    
    pad = (16,5) # padding around button text
    border = 2 # border around button text
    
    # font.render(text, antialias, text_color, background)
    # "Antialiasing" removes jagged outlines to text with some sub-pixel calc.
    # these render text to a new surface, then return that surface.
    # default background is "None" which would be transparent. Otherwise,
    #   that argument takes a color.

    # get centerx position for surface
    cx = state.surface.get_rect().centerx
    
    # create the title element
    self.title = UIText(fg=fg, bg=bg, font=self.title_font,
                        text="Break It!", y=32, centerx=cx)
    self.credit = UIText(fg=fg, bg=bg, font=self.credit_font,
                         text="by Aubrey Smith", y=128, centerx=cx)

    # create the Play and Quit buttons
    self.play_button = UIButton(fg=fg, bg=bg, font=self.button_font,
                                text="Play", y=275, centerx=cx,
                                padding=pad, border=border, bordercolor=fg)
    self.quit_button = UIButton(fg=fg, bg=bg, font=self.button_font,
                                text="Quit", y=350, centerx=cx,
                                padding=pad, border=border, bordercolor=fg)

  # render() renders the menu to a surface.
  def render(self, surface):
    # first, fill the entire window surface with our background color.
    surface.fill(self.bg)
    
    # blit the title and buttons to their positions.
    # if we included a third argument it would let us use the .surfrect
    #   for the element; since we're using the whole surface for each
    #   element, we don't need that argument.
    self.title.render(surface)
    self.credit.render(surface)
    self.play_button.render(surface)
    self.quit_button.render(surface)
    # and that's it! Simple enough.
    
  # Should be sent a MOUSEBUTTONDOWN event. Tests for clicks on buttons.
  # returns True if a click landed inside a button.
  def mousedown(self, event):
    # grab event pos
    pos = event.pos
    
    # check collision of UI button rects with mouse position
    # if the function returns True, the player has mouse down on that button.
    if self.play_button.rect.collidepoint(pos):
      self.clicking = self.play_button
    elif self.quit_button.rect.collidepoint(pos):
      self.clicking = self.quit_button
    else:
      self.clicking = None
    if self.clicking == None:
      return False
    else:
      return True
  
  # Should be sent a MOUSEBUTTONUP event. Tests for clicks on buttons.
  # return True if a click has been completed on a button
  # (the same as .moused
  def mouseup(self, event):
    # if we're not currently clicking on anything, this just fails.
    if self.clicking == None:
      return False
    
    # grab event pos
    pos = event.pos
    
    # check collision of UI button rects with mouse position
    # if the function returns True, the player has mouse down on that button.
    if self.clicking != None and self.clicking.rect.collidepoint(pos):
      return True
    else:
      self.clicking = None
      return False

  def update(self, state):
    # main menu event loop
    # returns False if user quits
    # changes the state.screen if needed
    if state.need_update:
      # render the main menu and flip so the screen updates.
      self.render(state.surface)
      pygame.display.flip()
      state.need_update = False # until next we need an update.
    for e in pygame.event.get():
      if e.type == QUIT:
        state.screen = None
        return False # respond to quit signal
      elif e.type == MOUSEBUTTONDOWN:
        self.mousedown(e)
      elif e.type == MOUSEBUTTONUP:
        if self.mouseup(e):
          # means we have a successful click!
          # find out what got clicked and clear it.
          if self.clicking == self.play_button:
            # switch to play mode
            state.screen = state.playscreen
            state.playscreen.reset_game()
            state.need_update = True # because the screen needs to change
            self.clicking = None
          elif self.clicking == self.quit_button:
            # time to quit, then
            state.screen = None
            return False
    return True

class PlayScreen:
  def __init__(self, state):
    self.font = pygame.freetype.Font("font/OverpassMono/OverpassMono-SemiBold.ttf", 32)
    # make a back surface for us to work with;
    # it will be a fixed size.
    rect = pygame.Rect(0, 0, 320, 240)
    bigrect = state.surface.get_rect()
    self.surface = pygame.Surface(rect.size) # surface for play objects / sprites
    self.textsurf = pygame.Surface(bigrect.size,flags=SRCALPHA) # surface for text objects
    # set up our foreground and background elements
    fg = (255,255,255) # white for foreground color
    bg = (0,0,0) # black for background color
    self.bg = bg
    self.fg = fg
    
    self.game = Game(self) # create the dormant game object
    # play area
    self.playarea = UIElement(fg=fg, bg=bg, offset=self.game.playrect.topleft,
                              size=self.game.playrect.size,
                              bordercolor=fg, border=4)
    self.playarea.pad() # because raw UIElements don't know to pad themselves

  def reset_game(self):
    self.game = Game(self)
    # would load first level here.
    self.timeout = 180

  def render(self, surface):
      # display text at high res and sprites at low res.
      # blank our surfaces; sprite surface to bg, text to transparent.
      self.surface.fill(self.bg)
      self.textsurf.fill((0,0,0,0))
      # render the play area
      self.playarea.render(self.surface)
      # render everything else
      self.game.render(self.surface, self.textsurf)
      # is it game over / did we win? Display a message to that effect!
      if self.game.over or self.game.won:
        center = (self.game.playrect.centerx * 2, self.game.playrect.centery * 2)
        text = "ERROR"
        if self.game.over:
          text = "GAME OVER"
        elif self.game.won:
          text = "YOU WON!"
        UIText(fg=self.fg, bg=self.bg, center=center,
               font=self.font, text=text).render(self.textsurf)
      # render the play area onto the main surface
      pygame.transform.scale(self.surface, surface.get_rect().size, surface)
      # render text on top of that
      surface.blit(self.textsurf, (0,0))
    
  def update(self, state):
    for e in pygame.event.get():
      if e.type == QUIT:
        state.screen = None
        return False # respond to quit signal
      elif e.type == KEYDOWN:
        # send key events to the paddle first.
        if self.game.input(e) == False:
          if e.key == K_ESCAPE:
            # for now, escapes to main menu.
            state.screen = state.mainmenu
            state.need_update = True
      elif e.type == KEYUP:
        self.game.input(e) # don't care about return value here.
    if self.game.over or self.game.won:
      # check for timeout; return to main menu when over.
      self.timeout -= 1
      if self.timeout <= 0:
        state.screen = state.mainmenu
        state.need_update = True
    else:
      # continue
      self.game.step()
    # post new frame to screen if needed
    if state.screen == self and state.need_update:
      self.render(state.surface)
      pygame.display.flip()
      if(self.game.over or self.game.won):
        state.need_update = False # Game Over screens are static, so don't re-render
    return True


class Paddle:
  def __init__(self, game):
    self.game = game
    self.rect = pygame.Rect((0,0),(32,8))
    self.sprite = game.spritesheet.sprites["paddle"]
    self.reset_pos()
    self.v = [0,0] # current paddle velocity, y will always be 0
    self.held = 0 # frames a movement key is held
    # directional key down? set it True here.
    self.l = False
    self.r = False

  def reset_pos(self):
    # resets the position of the paddle to the middle of the screen
    self.rect.centerx = self.game.playrect.centerx
    self.rect.bottom = self.game.playrect.bottom - 8
  
  def hit(self, obj):
    # handles paddle being hit by other mobile objects
    if isinstance(obj, Ball):
      # ball hit the paddle, apply english
      # note: if catch powerup is active, this is where we should check & catch
      #   instead of bouncing!
      (cx, cy, t) = whencollide(self, obj) # get time & planes of collision
      v = obj.v
      rect = obj.rect
      rect.x += int(t * v[0])
      rect.y += int(t * v[1])
      angle = self.angle(obj) # gives new x velocity float according to paddle v & ball v
      vp = v[:] # copy v
      vp[0] = int(angle * abs(v[1]))
      if v[1] > 0:
        vp[1] = -v[1]
      rect.x += int((1 + t) * vp[0])
      rect.y += int((1 + t) * vp[1])
      obj.v = vp
    else:
      print("Unhandled object hit paddle, type", type(obj))

  def angle(self, ball):
    # determines angle to fly ball at, according to how close the ball is to the edge of
    # the paddle. Returns the relative x velocity of the ball.
    f = int((ball.rect.centerx - self.rect.centerx) * 32 / self.rect.w)
    if f < 0:
      sign = -1
    else:
      sign = 1
    if abs(f) < 10:
      a = sign * 0.5
    elif abs(f) < 14:
      a = sign
    elif abs(f) < 16:
      a = 1.5 * sign
    else:
      a = 2 * sign
    return a

  def launch(self):
    # launches the ball(s) attached to the paddle
    for b in self.game.balls:
      if b.caught == True:
        b.v[1] = -self.game.ballinitvy
        b.v[0] = self.v[0]
        if abs(b.v[0]) < self.game.minballvx:
          if b.v[0] > 0:
            b.v[0] = self.game.minballvx
          else:
            b.v[0] = -self.game.minballvx
        b.caught = False
  
  def step(self):
    # moves time forward on paddle by one step
    if self.l and self.r:
      # both buttons down, do not move
      self.held = 0
      self.v[0] = 0
      return
    elif self.l:
      # move left
      if self.held >= self.game.accelframes:
        self.v[0] = -self.game.fastpv
      else:
        self.v[0] = -self.game.slowpv
      self.held += 1
    elif self.r:
      # move right
      if self.held >= self.game.accelframes:
        self.v[0] = self.game.fastpv
      else:
        self.v[0] = self.game.slowpv
      self.held += 1
    else:
      # no buttons held, do not move
      self.held = 0
      self.v[0] = 0
      return

    oldx = self.rect.x
    self.rect.x += self.v[0]
    if self.rect.left < self.game.playrect.left:
      self.rect.left = self.game.playrect.left
    elif self.rect.right > self.game.playrect.right:
      self.rect.right = self.game.playrect.right
    dx = self.rect.x - oldx
    for b in self.game.balls:
      # move every ball stuck to the paddle along with the paddle.
      if b.caught:
        b.rect.x += dx

  def render(self, surface):
    self.sprite.render(surface, self.rect.topleft)

  def input(self, e):
    # takes an input event (key event) and determines if it's actionable by the paddle.
    # left/a, right/d keys move the paddle; holding accelerates to a doubled speed.
    # w/up/spacebar shoot the ball, if it's caught on the paddle.
    if e.type == KEYDOWN:
      if e.key == K_a or e.key == K_LEFT:
        self.l = True
      elif e.key == K_d or e.key == K_RIGHT:
        self.r = True
      elif e.key == K_w or e.key == K_UP or e.key == K_SPACE:
        self.launch() # launches the ball(s) attached to the paddle
      else:
        return False
    elif e.type == KEYUP:
      if e.key == K_a or e.key == K_LEFT:
        self.l = False
      elif e.key == K_d or e.key == K_RIGHT:
        self.r = False
    else:
      return False
    return True

# class for the ball that does the actual brick-breaking
class Ball:
  def __init__(self, game, pos=None, v=None):
    # pos and v are useful when spawning new balls, say, for multiball
    if v != None:
      self.v = list(v)
    else:
      self.v = [0,0]
    if pos == None:
      rect = game.paddle.rect
      pos = (rect.centerx - 3, rect.top - 6)
      self.caught = True
    else:
      self.caught = False
    self.rect = pygame.Rect(pos, (6,6))
    self.sprite = game.spritesheet.sprites["ball"]
    self.game = game
    self.dead = False

  def render(self, surface):
    self.sprite.render(surface, self.rect.topleft)

  def step(self):
    if self.caught or self.dead:
      # no physics to do if ball is stuck to paddle or is dead
      return
    prect = self.game.playrect
    v = self.v
    self.rect.x += v[0]
    self.rect.y += v[1]
    rect = self.rect
    # check collision with floor -- loses ball
    if rect.bottom > prect.bottom:
      self.dead = True
      self.game.destroy_ball(self) # kills ball immediately; no animation
      return
    # check collisions with walls
    if rect.left < prect.left:
      # bounce off
      v[0] = -v[0]
      rect.x += (prect.left - rect.left) * 2
    elif rect.right > prect.right:
      # bounce off
      v[0] = -v[0]
      rect.x -= (rect.right - prect.right) * 2
    if rect.top < prect.top:
      # bounce off
      v[1] = -v[1]
      rect.y += (prect.top - rect.top) * 2
    # check collision with paddle
    if rect.colliderect(self.game.paddle.rect):
      self.game.paddle.hit(self)
    # check collision with bricks
    for b in self.game.curlvl.bricks:
      if rect.colliderect(b.rect):
        b.hit(self)

# Timer class counting seconds:frames assuming 60 frames per second.
class Timer:
  def __init__(self, screen, s=0, f=0, offset=(0,0), center=None, font=None):
    # the above are for fixing the starting time, and our text's position
    # in the appropriate UI co-ordinate system.
    self.done = False # set True when timer hits 0
    self.frames = s * 60 + f # assumes 60 fps
    if(center == None):
      self.uitext = UIText(fg=screen.fg, bg=screen.bg, font=font, offset=offset)
    else:
      self.uitext = UIText(fg=screen.fg, bg=screen.bg, font=font, center=center)
    self.refresh_text()

  def refresh_text(self):
    # sets the text to the appropriate contents
    # assumes we want zero-padding for both segments, and 2 digits each.
    s = self.frames // 60
    f = self.frames % 60
    self.uitext.set_text("{:02}:{:02}".format(s,f))

  def step(self):
    # moves time forward one tick, subtracting 1 from the timer frames
    # unless timer is already at 0, in which case we stop.
    if self.frames <= 0:
      return
    self.frames -= 1
    if self.frames <= 0:
      self.done = True
    self.refresh_text()

  def render(self, surface):
    # pass-through render to our UIText
    self.uitext.render(surface)

# Brick object; represents a brick. Has a position in the play area, and will
# collide with the ball to break on contact (the hit() method)
class Brick:
  def __init__(self, game, index, sprite):
    # note: index is the level data index, not the brick's list index.
    # levels are always 11 wide, in this setup.
    self.game = game
    x = (index % 11) * 16 + game.playrect.x
    y = (index // 11) * 8 + game.playrect.y
    self.rect = pygame.Rect((x,y), (16, 8))
    self.sprite = sprite
    self.hp = 1
    self.pickup = None

  def render(self, surface):
    self.sprite.render(surface, self.rect.topleft)

  def hit(self, obj):
    # handles being hit by the ball; could also handle being hit by things
    # other than balls, such as laser shots. Will only cause balls to bounce.
    if isinstance(obj, Ball):
      # balls need to reflect!
      (cx, cy, t) = whencollide(self, obj) # get time & planes of collision
      v = obj.v
      vp = list(v)
      if cx:
        vp[0] = -v[0]
      if cy:
        vp[1] = -v[1]
      # perform the reflection calculation.
      obj.rect.x += int(t * v[0] + (1 + t) * vp[0])
      obj.rect.y += int(t * v[1] + (1 + t) * vp[1])
      obj.v = vp
    # regardless, damage the brick
    self.hp -= 1 # if we include things that do more than 1 damage, check here.
    if self.hp <= 0:
      if self.pickup != None:
        self.game.add_pickup(self.pickup, center=self.rect.center)
      self.game.destroy_brick(self)

# Generic Pickup class: covers the general behavior of pickups.
# Extensions should have a name beginning with Pup, in CamelCase.
# Extensions should implement the "hit(self, obj)" method.
class Pickup:
  def __init__(self, game):
    self.game = game
    self.rect = None
    self.v = (0,1) # pickups float downward, typically.
    self.sprite = None # extending Pups should set this sprite.

  def render(self, surface):
    self.sprite.render(surface, self.rect.topleft)

  def step(self):
    # moves time forward by one step.
    prect = self.game.playrect
    v = self.v
    self.rect.x += v[0]
    self.rect.y += v[1]
    rect = self.rect
    # check collision with floor -- makes pickup disappear
    if rect.bottom > prect.bottom:
      self.dead = True
      self.game.destroy_pickup(self) # kills pickup immediately; no animation
      return
    # check collision with paddle
    elif rect.colliderect(self.game.paddle.rect):
      self.hit(self.game.paddle)
      self.game.destroy_pickup(self)

  def hit(self,obj):
    pass # dummy function; implement in extension Pups.

class PupExtraLife(Pickup):
  def __init__(self,game):
    super().__init__(game)
    self.sprite = game.spritesheet.sprites["p-pup"]

  def hit(self, obj):
    # give the player an extra life!
    self.game.add_life()

# Level object loads the level, complete with all bricks and power-ups.
# (power-ups are not yet implemented.)
class Level:
  def __init__(self, game, data, powerups=[]):
    self.game = game
    # data must be a single, long string, with no whitespace.
    self.data = data
    # powerups is a list made of triplets. Each goes string, x, y.
    # The string is a single letter, lower case. The only one
    # implemented so far is p.
    self.powerups = powerups
    # generate empty lists for breakables and physical pickups
    self.bricks = []
    self.bcount = 0
    self.pickups = []
    self.pcount = 0
    self.won = False # true when level is won (no need to keep the ball moving)
    self.end = False # true when level ends normally.

  def load(self):
    # loads the level from the data we were given before.
    self.bricks = []
    grid = {}
    self.pickups = []
    self.bcount = self.pcount = 0
    game = self.game
    sprites = self.game.spritesheet.sprites

    i = 0
    # place bricks
    for b in self.data:
      if b == ".":
        pass # do nothing on . as it's a blank space
      elif b == "b":
        # blue brick
        br = Brick(game, i, sprites['b-brick'])
        self.bricks.append(br)
        grid[i] = br
        self.bcount += 1
      elif b == "g":
        # green brick
        br = Brick(game, i, sprites['g-brick'])
        self.bricks.append(br)
        grid[i] = br
        self.bcount += 1
      elif b == "o":
        # orange brick
        br = Brick(game, i, sprites['o-brick'])
        self.bricks.append(br)
        grid[i] = br
        self.bcount += 1
      elif b == "p":
        # purple brick
        br = Brick(game, i, sprites['p-brick'])
        self.bricks.append(br)
        grid[i] = br
        self.bcount += 1
      elif b == "r":
        # red brick
        br = Brick(game, i, sprites['r-brick'])
        self.bricks.append(br)
        grid[i] = br
        self.bcount += 1
      elif b == "y":
        # yellow brick
        br = Brick(game, i, sprites['y-brick'])
        self.bricks.append(br)
        grid[i] = br
        self.bcount += 1
      else:
        raise Exception("Level.load() received data with improper characters, closing.")
      
      i += 1

    # place powerups in bricks
    i = 0
    while i < len(self.powerups):
      p = self.powerups[i]
      x = int(self.powerups[i+1])
      y = int(self.powerups[i+2])
      j = x + y * 11
      if p == "p":
        grid[j].pickup = PupExtraLife(game)
      else:
        # non-implemented pickup
        raise Exception("Level.load() received pickup data with improper characters, closing.")
      i += 3

  def render(self, surface):
    for b in self.bricks:
      b.render(surface)
    for p in self.pickups:
      p.render(surface)

  def destroy_brick(self, brick):
    i = self.bricks.index(brick)
    del(self.bricks[i])
    self.bcount -= 1
    if self.bcount <= 0:
      self.won = True
      if self.pcount <= 0:
        self.end = True
        
  def add_pickup(self, pickup, offset=(0,0), center=None):
    rect = pickup.sprite.rect.copy()
    if center != None:
      rect.center = center
    else:
      rect.topleft = offset
    pickup.rect = rect
    self.pcount += 1
    self.pickups.append(pickup)

  def destroy_pickup(self, pickup):
    i = self.pickups.index(pickup)
    del(self.pickups[i])
    self.pcount -= 1
    if self.bcount <= 0 and self.pcount <= 0:
      self.end = True

# game state object; handles the state of play for the game. Initialize a new
# Game state every time the PlayScreen is re-opened.
class Game:
  def __init__(self, screen):
    self.screen = screen
    self.playrect = pygame.Rect((16,10), (176,220))
    # load sprite sheet
    self.spritesheet = SpriteSheet("img/sprites.png",
     [0,0,8,8,    #  0: (P)
      8,0,8,8,    #  1: (E)
      16,0,8,8,   #  2: (C)
      24,0,8,8,   #  3: (L)
      32,0,8,8,   #  4: (F)
      40,0,8,8,   #  5: (M)
      0,8,32,8,   #  6: Paddle
      32,8,6,6,   #  7: Ball
      40,8,6,8,   #  8: Laser shot?
      0,16,16,8,  #  9: red brick
      16,16,16,8, # 10: orange brick
      32,16,16,8, # 11: yellow brick
      0,24,16,8,  # 12: green brick
      16,24,16,8, # 13: blue brick
      32,24,16,8, # 14: purple brick
      "p-pup",0,
      "e-pup",1,
      "c-pup",2,
      "l-pup",3,
      "f-pup",4,
      "m-pup",5,
      "paddle",6,
      "ball",7,
      "laser",8,
      "r-brick",9,
      "o-brick",10,
      "y-brick",11,
      "g-brick",12,
      "b-brick",13,
      "p-brick",14])
    # load basic items
    self.paddle = Paddle(self)
    # balls are using a list for future-proofing (multiball)
    self.balls = [Ball(self)]
    self.over = False # True when Game Over
    self.won = False # True when You Won!
    self.waitlvl = False # true when waiting for next level
    self.waitlife = False # true when waiting after losing a paddle
    self.wait = 60 # storage for frame counter when waiting for next level
    self.maxballvx = 8
    self.minballvx = 1
    self.ballinitvy = 2
    self.fastpv = 4
    self.slowpv = 2
    self.accelframes = 6
    # load levels
    self.lvl = 1
    self.levels = []
    self.levels.append(Level(self,
      "..........." * 3 +
      "bogrypyrgob" +
      "ogrypppyrgo" +
      "grypp.ppyrg",
      powerups=['p',5,3]))
    self.levels.append(Level(self,
      "..........." * 2 +
      ".....b....." +
      "....byb...." +
      "..g..b..g.." +
      ".gog...gog." +
      "gopog.gopog" +
      ".gog...gog." +
      "..g..b..g.." +
      "....byb...." +
      ".....b.....",
      powerups=['p',5,3,
                'p',2,6,
                'p',8,6,
                'p',5,9]))
    self.curlvl = self.levels[self.lvl - 1]
    self.curlvl.load()
    # level indicator
    font = pygame.freetype.Font("font/OverpassMono/OverpassMono-SemiBold.ttf", 32)
    self.lvltext = UIText(fg=screen.fg, bg=screen.bg,
                          font=font, offset=(432,128), text="Level " + str(self.lvl))
    # life count and indicator
    self.lives = 3
    self.livestext = UIText(fg=screen.fg, bg=screen.bg,
                          font=font, offset=(432,160), text="Lives: " + str(self.lives))

  def step(self):
    # steps time forward one frame
    for b in self.balls:
      b.step()
    for p in self.curlvl.pickups:
      p.step()
    if self.paddle != None:
      self.paddle.step()
    if self.waitlvl:
      self.wait -= 1
      if self.wait <= 0:
        self.next_level()
    elif self.waitlife:
      self.wait -= 1
      if self.wait <= 0:
        self.next_life()

  def render(self, spritesurf, textsurf):
    # the two surfaces are for the expanded sprite surface vs. smooth text surface
    for b in self.balls:
      b.render(spritesurf)
    if self.paddle != None:
      self.paddle.render(spritesurf)
    self.curlvl.render(spritesurf)
    self.lvltext.render(textsurf)
    self.livestext.render(textsurf)

  def input(self, e):
    # takes events, passes to paddle if we have one
    if self.paddle != None:
      return self.paddle.input(e)
    return False

  def destroy_ball(self, ball):
    try:
      b = self.balls.index(ball)
      del self.balls[b]
      if(self.won != True and self.curlvl.won != True
          and len(self.balls) == 0):
        # lose a life
        self.wait_life()
    except ValueError:
      print("Game.destroy_ball() received ball not in current list; ignoring.")

  def destroy_brick(self, brick):
    self.curlvl.destroy_brick(brick)
    if self.curlvl.won:
      # destroy all balls, we don't need them anymore
      for b in self.balls:
        self.destroy_ball(b)
    if self.curlvl.end:
      self.wait_level() # set up for next level.

  def add_pickup(self, pickup, offset=(0,0), center=None):
    self.curlvl.add_pickup(pickup, offset, center)

  def destroy_pickup(self, pickup):
    self.curlvl.destroy_pickup(pickup)
    if self.curlvl.end:
      self.wait_level() # set up for next level.

  def wait_level(self):
    self.wait = 60
    self.waitlvl = True

  def next_level(self):
    self.waitlvl = False
    self.levels[self.lvl-1] = None
    self.lvl += 1
    if self.lvl <= len(self.levels):
      self.curlvl = self.levels[self.lvl-1]
      self.curlvl.load()
      if self.paddle == None:
        self.paddle = Paddle(self)
      self.paddle.reset_pos() # reset paddle to center
      self.balls = [Ball(self)]
      self.lvltext.set_text("Level " + str(self.lvl))
    else:
      self.won = True # game is won when all levels finished!

  def add_life(self):
    # add one life
    self.lives += 1
    self.livestext.set_text("Lives: " + str(self.lives))

  def wait_life(self):
    # lose one life and set up a short wait for the next
    self.waitlife = True
    self.wait = 60
    self.lives -= 1
    self.livestext.set_text("Lives: " + str(self.lives))
    self.paddle = None

  def next_life(self):
    self.waitlife = False
    if self.lives <= 0:
      self.over = True # game over when all lives lost
      return # no point in doing the rest of this if game over
    # otherwise, load up the paddle again!
    self.paddle = Paddle(self)
    self.balls = [Ball(self)]

class State:
  # state information for entire game engine; basically a container to prevent
  # things spilling into the global namespace
  def __init__(self):
    self.screen = None
    # create a window that's 640 wide by 480 tall in pixels
    self.surface = pygame.display.set_mode((640,480))
    # give it a title
    pygame.display.set_caption("Break It!")
    self.mainmenu = MainMenu(self)
    self.playscreen = PlayScreen(self)
    self.need_update = True

    self.clock = pygame.time.Clock()

# init and main loop
def main():
  state = State()
  state.screen = state.mainmenu
  while(state.screen != None):
    state.screen.update(state) # handle all events, rendering, and screen switches
    state.clock.tick(60) # ensure loop runs no faster than 60 FPS

main()

pygame.quit() # clean up after pygame.
