# Engine code for Break It! example game for Jamming with Python
# programming and computer science history course content.
# (C) Lotus Fox LLC, Aubrey Smith, 2017

import pygame
from pygame.locals import *
import pygame.freetype


# A user interface element, abstract base class.
# Handles a bunch of layout information:
#   offset would take a tuple of co-ordinates, which can be replaced with
#    x/y, or with center or centerx/y, for setting either the top left
#    corner (for offset parts) or the center. x/y versions take single ints.
#   padding takes either a single number or a tuple for x/y padding. The single
#    number is treated as both numbers in the tuple otherwise.
#   fg, bg and bordercolor take pygame colors; these can be Colors or tuples
#    in (r,g,b) or (r,g,b,a) format. They correspond to foreground/text color,
#    background color, and border color, respectively.
#   border is an integer representing the thickness of the border around the
#    object.
#   Most UI elements will need this offset.
# Stores the surface for its appearance, to be rendered by another object.
# Also stores the rect defining its position in the UI, for collision detection.
class UIElement:
  rect = None
  padding = (0,0)
  
  def __init__(self,
               offset=(0,0), x=None, y=None,
               center=None, centerx=None, centery=None,
               size=(0,0), w=None, h=None,
               padding=0, fg=(255,255,255), bg=None,
               bordercolor=(255,255,255), border=0):
    
    # calculate rect in window from our size info.
    dim = list(size)
    if w != None:
      dim[0] = w
    if h != None:
      dim[1] = h
    self.rect = pygame.Rect(offset, dim)

    # change offset if x or y params are set
    if x != None:
      self.rect.x = x
    if y != None:
      self.rect.y = y
    
    # change center, or axis-wise centers, as requested.
    self.centered = [False,False]
    if center != None:
      self.rect.center = center
      self.centered = [True,True]
    else:
      if centerx != None:
        self.rect.centerx = centerx
        self.centered[0] = True
      if centery != None:
        self.rect.centery = centery
        self.centered[1] = True

    if isinstance(padding, int):
      self.padding = (padding,padding)
    else:
      self.padding = padding
    self.fg = fg
    self.bg = bg
    self.bordercolor = bordercolor
    self.border = border

  # use this function to retrieve the offset, e.g. when blitting.
  def get_offset(self):
    return (self.rect.x, self.rect.y)

  # functions to set offset or center, affecting centered behavior
  def set_offset(self, offset):
    self.rect.topleft = offset
    self.centered = [False,False]
  def set_x(self, x):
    self.rect.x = x
    self.centered[0] = False
  def set_y(self, y):
    self.rect.y = y
    self.centered[1] = False
  def set_center(self, center):
    self.rect.center = center
    self.centered = [True,True]
  def set_centerx(self, cx):
    self.rect.centerx = cx
    self.centered[0] = True
  def set_centery(self, cy):
    self.rect.centery = cy
    self.centered[1] = True

  # functions to set size, or width or height individually
  def set_size(self, size):
    center = list(self.rect.center)
    offset = list(self.rect.topleft)
    self.rect.size = size
    if self.centered[0]:
      self.rect.centerx = center[0]
    else:
      self.rect.x = offset[0]
    if self.centered[1]:
      self.rect.centery = center[1]
    else:
      self.rect.y = offset[1]
  def set_w(self,w):
    cx = self.rect.centerx
    ox = self.rect.x
    self.rect.w = w
    if self.centered[0]:
      self.rect.centerx = cx
    else:
      self.rect.x = ox
  def set_width(self,w):
    self.set_w(w)
  def set_h(self,h):
    cy = self.rect.centery
    oy = self.rect.y
    self.rect.h = h
    if self.centered[1]:
      self.rect.centery = cy
    else:
      self.rect.y = oy
  def set_height(self,h):
    self.set_h(h)

  def pad(self):
    # internal function; sets size according to padding and rectangle
    # should be called by extending objects when they change content
    adds = list(self.padding)
    adds[0] += self.border
    adds[1] += self.border
    self.set_size((self.rect.w + 2 *adds[0], self.rect.h + 2 *adds[1]))
    if self.centered[0] != True:
      self.rect.x -= adds[0]
    if self.centered[1] != True:
      self.rect.y -= adds[1]

  def get_content_offset(self):
    # gets the offset for the contents of the object, in the same co-ordinates
    # as the object rect itself.
    adds = list(self.padding)
    adds[0] += self.border
    adds[1] += self.border
    return (self.rect.x + adds[0], self.rect.y + adds[1])
  
  def draw_border(self, surface):
    # draws border with border-color and thickness inside element rect
    # we have to fill a rect for every border.
    if self.border == 0:
      return
    
    rect = self.rect
    lines = [rect.copy(), rect.copy(), rect.copy(), rect.copy()]
    # [top, left, bottom, right]
    lines[0].h = lines[2].h = self.border
    lines[1].w = lines[3].w = self.border
    lines[2].bottom = rect.bottom
    lines[3].right = rect.right
    for l in lines:
      # fill a rect for every border; we set this up above
      pygame.draw.rect(surface, self.bordercolor, l)

  def render(self, surface):
    # draws background and border; all UIElement sub-classes should run this
    # before their own render function
    if self.bg != None:
      surface.fill(self.bg, self.rect)
    self.draw_border(surface)

# Visible text for UI.
# in addition to UIElement parts, takes:
#  font: the pygame.freetype.Font or font file; uses default if left blank
#  fontsize: the nominal size of the font (pixels per font height / "em size")
#    will be 16 if left blank
#  text: the actual text to display.
class UIText(UIElement):
  font = None
  text = ""

  def __init__(self,
               offset=(0,0), x=None, y=None,
               center=None, centerx=None, centery=None,
               size=(0,0), w=None, h=None,
               padding=0, fg=(255,255,255), bg=None,
               bordercolor=(255,255,255), border=0,
               font=None, fontsize=16, text=""):
    
    super().__init__(offset, x, y, center, centerx, centery, size, w, h,
                     padding, fg, bg, bordercolor, border)
    try:
      self.set_font(font)
    except TypeError:
      raise TypeError("Wrong type sent to UIText() for font, must be string path or pygame.freetype.Font")

    self.text = text
    self.offset = (0,0)
    self.refresh_size()

  def render(self, surface):
    super().render(surface)
    # render our text
    os = self.get_content_offset()
    self.font.render_to(surface, (self.offset[0] + os[0], self.offset[1]+os[1]), self.text, self.fg)

  def refresh_size(self):
    rect = self.font.get_rect(self.text)
    asc = self.font.get_sized_ascender()
    dsc = self.font.get_sized_descender()
    self.offset = (rect.x, asc-rect.y)
    rect.h = asc-dsc
    rect.w += rect.x
    self.set_size(rect.size)
    self.pad()

  def set_text(self, text):
    if self.text == text:
      return
    self.text = text
    self.refresh_size()

  def set_font(self, font):
    if isinstance(font, str):
      # need to load font from file
      self.font = pygame.freetype.Font(font, fontsize)
    elif font == None:
      # default font
      self.font = pygame.freetype.Font(None)
    else:
      # hope it's a pygame.freetype.Font
      if isinstance(font, pygame.freetype.Font):
        self.font = font
      else:
        print("Somehow, this is a", type(font))
        raise TypeError("Wrong type sent to UIText.font(), must be string path or pygame.freetype.Font")
    self.refresh_size()

# button class that uses visible text; typically should have a border to denote
#   that this is a button and not just text.
# onclick: the function to perform when the button is clicked. Should be sent
#   self, plus the Event and the State of the engine.
# hoverbg/fg: color of background and foreground when hovering over object
# hoverbc: hovering border color
# activebg/fg/bc: same as above but for when button is pressed
class UIButton(UIText):
  def __init__(self,
               offset=(0,0), x=None, y=None,
               center=None, centerx=None, centery=None,
               size=(0,0), w=None, h=None,
               padding=0, fg=(255,255,255), bg=None,
               bordercolor=(255,255,255), border=0,
               font=None, fontsize=16, text="",
               onclick=None, hoverbg=None, hoverfg=None, hoverbc=None,
               activebg=None, activefg=None, activebc=None):
    super().__init__(offset, x, y, center, centerx, centery, size, w, h,
                     padding, fg, bg, bordercolor, border, font, fontsize, text)
    self.onclick = onclick
    self.hoverbg = hoverbg
    self.hoverfg = hoverfg
    self.hoverbc = hoverbc
    self.activebg = activebg
    self.activefg = activefg
    self.activebc = activebc
    self.normalbg = bg
    self.normalfg = fg
    self.normalbc = bordercolor
    self.hovered = self.active = False

  def refresh_colors(self):
    # sets colors appropriately when hover / active state changes.
    self.bg = self.normalbg
    self.fg = self.normalfg
    self.bordercolor = self.normalbc
    
    if self.hovered:
      if self.hoverbg != None:
        self.bg = self.hoverbg
      if self.hoverfg != None:
        self.fg = self.hoverfg
      if self.hoverbc != None:
        self.bordercolor = self.hoverbc
    if self.active:
      if self.activebg != None:
        self.bg = self.hoverbg
      if self.activefg != None:
        self.fg = self.hoverfg
      if self.activebc != None:
        self.bordercolor = self.hoverbc

  def hover(self):
    # sent when hovering over the object
    self.hovered = True
    self.refresh_colors()
  def unhover(self):
    # sent when hover leaves the object
    self.hovered = False
    self.refresh_colors()
  def activate(self):
    # sent when button is pressed
    self.active = True
    self.refresh_colors()
  def deactivate(self):
    # sent when button is no longer pressed
    self.active = False
    self.refresh_colors()

  def inside(self, point):
    # tests if a point is inside of the button
    return self.rect.collidepoint(point)

# Collection for making Sprites out of a single file.
class SpriteSheet:
  def __init__(self, file, stream):
    # takes a filename for the image, and a stream for the sprite indexing.
    # stream is a list of strings; quartets denote rectangles.
    # string followed by number denotes a sprite name for sprites dict.
    self.image = pygame.image.load(file)
    self.slist = []
    self.sprites = {}
    i = 0
    while i < len(stream):
      if isinstance(stream[i], int) or stream[i].isdigit():
        # must be the start of a quartet, pull four out
        x = int(stream[i])
        y = int(stream[i+1])
        w = int(stream[i+2])
        h = int(stream[i+3])
        i += 4
        self.slist.append(Sprite(self, pygame.Rect(x,y,w,h)))
      else:
        # must be a name followed by an index
        self.sprites[stream[i]] = self.slist[int(stream[i+1])]
        i += 2

# a single Sprite out of a SpriteSheet; has a rect taken out of the sheet.
# will render() only itself out of that sheet when asked.
class Sprite:
  def __init__(self, sheet, rect):
    self.sheet = sheet
    self.rect = rect

  def render(self, surface, offset):
    # renders the sprite to the target
    surface.blit(self.sheet.image, offset, self.rect)

def whencollide(a, b):
  # checks collision time between two objects
  # assumes there is already a collision, so the time expected is in the past.
  # also assumes a is static and b is mobile (for simplicity)
  # returns a tuple of (bool, bool, float) for hit-x, hit-y, and time of collision.
  tx = ty = 0.0
  ar = a.rect
  br = b.rect
  hx = hy = False
  if b.v[0] > 0:
    tx = (ar.left - br.right) / b.v[0]
    if tx < 0:
      hx = True
  elif b.v[0] < 0:
    tx = (ar.right - br.left) / b.v[0]
    if tx < 0:
      hx = True
  if b.v[1] > 0:
    ty = (ar.top - br.bottom) / b.v[1]
    if ty < 0:
      hy = True
  elif b.v[1] < 0:
    ty = (ar.bottom - br.top) / b.v[1]
    if ty < 0:
      hy = True

  # if we have both planes colliding, we need to know which is the real collision.
  if hx == True and hy == True:
    # whichever apparently collision happened LATER is the correct one.
    if tx < ty:
      hx = False
    elif tx > ty:
      hy = False
    # otherwise leave both true, you somehow hit a corner!
  
  # finally, determine return value.
  if hx == hy:
    if tx < 0:
      # collided on a corner
      return (True, True, tx)
    else:
      # no collision
      return (False, False, 0.0)
  elif hx == True:
    return (hx, hy, tx)
  elif hy == True:
    return (hx, hy, ty)
  
  return (hx, hy, 0.0) # this line should never be reached

